#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <pthread.h>
#include <stdint.h>
#include <stdio.h>
#include <sys/time.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <assert.h>

#include "cornflake.h"

#define cpu_relax() __asm__ volatile("pause\n": : :"memory")

typedef uint32_t spinlock_t;

static inline int spin_trylock(spinlock_t *lock)
{
	size_t tries = 1000;

	while (tries--) {
		if (!__sync_lock_test_and_set(lock, 1))
			return 0;

		while (*lock) {
			cpu_relax();
		}
	}

	return -1;
}

static inline int spin_tryunlock(spinlock_t *lock)
{
	if (__sync_lock_test_and_set(lock, 0) == 0)
		return -1;

	return 0;
}

static inline void spin_unlock(spinlock_t *lock)
{
	__sync_lock_release(lock);
}

struct cornflake {
	spinlock_t lock;
	uint64_t last_time;
	uint16_t counter;
};

static uint8_t g_local_mac[6];

static struct cornflake *
cornflake_open(const char *path)
{
	int fd;
	struct cornflake *corn;

	fd = open(path, O_RDWR, 0666);
	if (fd < 0)
		return NULL;

	corn = mmap(NULL, sizeof(*corn), PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
	close(fd);

	return corn;
}

static struct cornflake *
cornflake_create(const char *path)
{
	int fd;
	struct cornflake *corn;

	fd = open(path, O_RDWR | O_CREAT | O_EXCL, 0666);
	if (fd < 0) {
		if (errno == EEXIST)
			return cornflake_open(path);

		return NULL;
	}

	if (ftruncate(fd, sizeof(*corn)) < 0) {
		close(fd);
		return NULL;
	}

	corn = mmap(NULL, sizeof(*corn), PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
	close(fd);

	corn->lock = 0;
	corn->last_time = 0;
	corn->counter = 0;

	return corn;
}

struct cornflake *
cornflake_new(const char *path)
{
	struct cornflake *corn;

	corn = cornflake_open(path);
	if (corn)
		return corn;

	return cornflake_create(path);
}

void
cornflake_close(struct cornflake *corn)
{
	munmap(corn, sizeof(*corn));
}

/**
 * Generate a 16-byte UID. The composition of the UID is as follows.
 *
 * [ 8 bytes .......][ 6 bytes ...][2 bytes]
 *  timestamp          macaddr      counter
 *
 *	Where:
 *
 *	- `timestamp` is the current time (in ms since the epoch) when the
 *	UID was generated
 *
 *	- `macaddr` is the MAC address for this computer, in raw form
 *
 *	- `counter` is a 16-bit counter to avoid collisions between UIDs
 *	generated in the same milisecond. This limits Cornflake to only
 *	2^16 UID/s per milisecond.
 *
 *	If successful, the `uid` buffer will be filled with the unique ID
 *	in raw form. Ensure that the buffer has 16 bytes before calling this
 *	function.
 *
 *	Return values:
 *
 *	- CORNFLAKE_OK (0): the UID was generated successfully
 *
 *	- CORNFLAKE_ELOCK: there was an error when trying to lock the global
 *	state for UID generation. This is most likely because another process
 *	crashed while keeping the lock. As part of the error handling mechanism,
 *	we will free the global lock, so further calls to `cornflake_gen` should
 *	be able to work.
 *
 *	- CORNFLAKE_EDRIFT: the system clock has drifted into the past. Cornflake
 *	will refuse to generate new UIDs until the clock reaches the time when the
 *	last UID was generated, as to ensure that no UIDs are tagged in the past.
 *
 *	- CORNFLAKE_EOVERFLOW: the counter in the UID has overflowed (there have been
 *	more than 2^16 unique IDs generated in this milisecond). You need to wait at
 *	least 1 milisecond and try again.
 */
int
cornflake_gen(uint8_t *uid, struct cornflake *corn)
{
	struct timeval tv;
	uint64_t timestamp;
	uint16_t sequence = 0;
	uint8_t *bytes;

	/*
	 * We need to synchronize the global state so only *one* Cornflake
	 * process generates UIDs simultaneously.
	 *
	 * We use a spinlock in shared memory (this requires x86/x64 semantics,
	 * do not try this at home, children, specially if you have an ARM
	 * processor at home).
	 *
	 * There is a caveat when locking this global state: there's an infintesimal
	 * chance that another Cornflake process got terminated (i.e. segmentation
	 * fault, uncaught signal) during the ~20 instructions that happen inside
	 * this lock, before the lock has been released.
	 *
	 * This would usually cause every other process in the system to deadlock
	 * while waiting for the spinlock. To work around this, we use a custom
	 * `trylock` the spinlock: we're going to try a thousand (1000) long locks
	 * in a row (i.e. try the exchange, if failed, loop until the lock is free,
	 * try the exchange again).
	 *
	 * If we haven't acquired a lock after 1000 tries, we'll assume the spinlock
	 * has deadlocked and we'll simply drop it and fail fast so other processes
	 * can continue UID generations. This all should happen in less than ~500ms
	 * even in the highest contention cases.
	 *
	 * FAQ: Why are you using an `xchg` spinlock instead of e.g. `pthread_mutex` on
	 * shared memory or `pthread_spinlock`? All the `pthread` synchronization
	 * primitives have undefined behavior when trying to unlock/destroy them from
	 * a different thread than the one who originally acquired the lock. By controlling
	 * the spinlock semantics, we can ensure the lock gets released when a process dies
	 * holding it *and* we can verify that other processes do not race on the released
	 * lock.
	 */
	if (spin_trylock(&corn->lock) < 0) {
		spin_unlock(&corn->lock);
		return CORNFLAKE_ELOCK;
	}

	/*
	 * Get the current time in ms -- this will be the timestamp that shows
	 * on the first 8 bytes of the UID
	 */
	gettimeofday(&tv, NULL);
	timestamp = ((uint64_t)tv.tv_sec * 1000) +
		((uint64_t)tv.tv_usec / 1000);

	/*
	 * If the current timestamp is behind the timestamp of the last UID we've
	 * generated, that means our system clock has drifted. We cannot generate
	 * new UIDs until the clock is back on time.
	 */
	if (timestamp < corn->last_time) {
		spin_unlock(&corn->lock);
		return CORNFLAKE_EDRIFT;
	}

	/*
	 * If the current timestamp is the same as the last UID's we've generated,
	 * we need to increase the UID counter (last 16 bits of the UID) to prevent
	 * collisions.
	 *
	 * When the UID counter wraps around (i.e. reaches 0), we've generated more
	 * than 2^16 UIDs in the millisecond, so we need to wait to the next ms
	 * in order to prevent collisions.
	 */
	if (timestamp == corn->last_time) {
		sequence = ++corn->counter;
		if (!sequence) {
			spin_unlock(&corn->lock);
			return CORNFLAKE_EOVERFLOW;
		}
	} else {
		corn->last_time = timestamp;
		corn->counter = 0;
	}

	/*
	 * We're out of the locking zone. HOWEVER: we need to ensure that the lock
	 * we've acquired is still acquired when we unlock it. If somehow the lock
	 * is not taken when we try to unlock it, that means another process has
	 * decided that we're "dead" (we took too long to generate this UID!) and
	 * unlocked the lock for us.
	 *
	 * In that case, we cannot continue: the UID we've just generated could very
	 * well be racy.
	 */
	if (spin_tryunlock(&corn->lock) < 0)
		return CORNFLAKE_ELOCK;

	/*
	 * Move over all the generated data to network byte order (big endian).
	 * This is to ensure that the UIDs are k-ordered when compared
	 * lexicographically.
	 */
	bytes = (uint8_t *)&timestamp;
	uid[0] = bytes[7];
	uid[1] = bytes[6];
	uid[2] = bytes[5];
	uid[3] = bytes[4];
	uid[4] = bytes[3];
	uid[5] = bytes[2];
	uid[6] = bytes[1];
	uid[7] = bytes[0];

	memcpy(&uid[8], g_local_mac, sizeof(g_local_mac));

	bytes = (uint8_t *)&sequence;
	uid[14] = bytes[1];
	uid[15] = bytes[0];

	return 0;
}

int
cornflake_gen_hex(char *hex_uid, size_t len, struct cornflake *corn)
{
	static char to_hex[] = "0123456789abcdef";

	uint8_t uid[16];
	int i, ret;
	
	if (len < 33)
		return CORNFLAKE_EOVERFLOW;

	ret = cornflake_gen(uid, corn);
	if (ret < 0)
		return ret;

	for (i = 0; i < 16; ++i) {
		*hex_uid++ = to_hex[uid[i] >> 4];
		*hex_uid++ = to_hex[uid[i] & 0xf];
	}

	*hex_uid = '\0';
	return 0;
}

int
cornflake_set_mac(const char *mac_addr)
{
	if (strlen(mac_addr) != 17)
		return CORNFLAKE_EMACADDR;

	if (sscanf(mac_addr, "%hhx:%hhx:%hhx:%hhx:%hhx:%hhx",
		&g_local_mac[0], &g_local_mac[1], &g_local_mac[2],
		&g_local_mac[3], &g_local_mac[4], &g_local_mac[5]) != 6)
		return CORNFLAKE_EMACADDR;

	return 0;
}

#ifdef CORNFLAKE_STANDALONE
#include "getopt.h"

static inline double bench_time(void)
{
	struct timespec ts;
	clock_gettime(CLOCK_REALTIME, &ts);
	return (ts.tv_sec * 1000.0) + (ts.tv_nsec / 1000000.0);
}


int main(int argc, char *argv[])
{
	static struct option longopts[] = {
		{ "count", required_argument, NULL, 'n' },
		{ "path", required_argument, NULL, 'o' },
		{ "time", no_argument, NULL,	't' },
		{ "version", no_argument, NULL,	'v' },
		{ NULL,  0,  NULL, 0 }
	};

	struct cornflake *flake;
	const char *flake_path = "./milkbowl";
	int opt, benchmark = 0;
	unsigned int i, count = 1;
	double start, end;

	while ((opt = getopt_long(argc, argv, ":n:o:tv", longopts, NULL)) != -1) {
		switch (opt) {
		case 'n': count = atoi(optarg); break;
		case 'o': flake_path = optarg; break;
		case 't': benchmark = 1; break;
		case 'v':
			puts("cornflake");
			return 0;

		default:
			printf("Usage: %s [--path CORNFLAKE_CONTEXT] [-n COUNT]\n", argv[0]);
			return 1;
		}
	}

	flake = cornflake_new(flake_path);
	if (flake == NULL) {
		fprintf(stderr, "Failed to open Cornflake context: %s\n", flake_path);
		return -1;
	}

	start = bench_time();

	for (i = 0; i < count; ++i) {
		char uid[33];
		int err = cornflake_gen_hex(uid, sizeof(uid), flake);

		if (err < 0) {
			fprintf(stderr, "Generation failed: %d\n", err);
			return -2;
		}

		puts(uid);
	}

	end = bench_time();

	if (benchmark)
		fprintf(stderr, "Benchmark: %d oids in %fms\n", count, end - start);

	cornflake_close(flake);
	return 0;
}
#endif
